//
//	Includes
//
#include	"MCF52233.h"	// Common headder for ColdFire
#include	"ProtoType.h"	// Prototype headder for SilentMoon system call

//
//	Prototypes
//
UINT8 SenseByte(void);	// Sense char of UART
UINT8 GetByte(void);		// Get char from UART
void	PutByte(UINT8);		// Put char for UART
void	InitDevice(UINT8 baud);	// Device init
void	SerialMain(void);	// Main loop
void	SetupData(void);		// Setup UDP packet
void	DataIndication(void);	// Handle incoming UDP packet
__declspec(interrupt) void TimerHandler(void);	// Timer handler
__declspec(interrupt) void SciHandler(void);		// UART handler

//
//	Constants (change by user)
//
#define	BAUD_128000	12
#define BAUD_115200	16
#define BAUD_57600	32
#define	BAUD_38400	49
#define	BAUD_28800	65
#define	BAUD_19200	97
#define	BAUD_14400	130
#define	BAUD_9600		195

static const UINT8 BaudRate[7] =	// Baud Rate table
{	BAUD_115200, BAUD_57600, BAUD_38400, BAUD_28800,
	BAUD_19200, BAUD_14400,	BAUD_9600};

//
//	Buffer size constant 
//
#define	ETHERBUF	255	// UDP payload size
#define	RX_SIZE		256	// UART receive buffer
#define	RX_MASK		0xff	// Mask for 1024 byte ring buffer
#define	TX_SIZE		256	// UART transmit buffer
#define	TX_MASK		0xff	// Mask for 1024 byte ring buffer

//
//	UDP Packet data struct
//
typedef	struct 
{
	UINT16	Size;		// Payload size
	UINT8	AckSeq;		// Acknowledged seq number
	UINT8	SeqNum;		// Packet seq number
	UINT8	Time;			// Arrival time

	UINT8	data[ETHERBUF];	// Data buffer
}	PacketBuffer_t;	// Typedef name of UDP packet buffer

//
//	Vars
//
UINT8	RxRingIn;		// Receive ring in pointer (UART)
UINT8	RxRingOut;	// Receive ring out pointer (UART)
UINT8	TxRingIn;		// Transmit ring in pointer (UART)
UINT8	TxRingOut;	// Transmit ring out pointer (UART)
UINT8	TxSeqNum;			// My packet seq number
UINT8	RxSeqNum;			// Received packet seq number
UINT8	RxLastNum;		// Last processed packet seq number
INT8 Soc;						// Socket handle for UDP send/receive
volatile UINT8 TickCount;	// Timer tick of 1ms timer
UINT8	Master;				// Master or Slave
UINT8	Adjust;				// Adjust time
UINT8	LoopTime;			// Main loop time
UINT8	Baud;					// Baud rate buffer
UINT16	SrcPort;		// Source port number
UINT16	DstPort;		// Destination port number
UINT32	ip;					// IP adrs

PacketBuffer_t	*TxPacketBuffer;	// Transmit ether packet buffer
PacketBuffer_t	*RxPacketBuffer;	// Recieve ether packet buffer
UINT8	*RxBuf;			// Receive ring buffer (UART)
UINT8	*TxBuf;			// Transmit ring buffer (UART)

//
//	Timer handler (1ms tick)
//
__declspec(interrupt) void TimerHandler(void)
{
	MCF_DTIM0_DTER = 2;	// Clear timer interrupt flag
	TickCount++;				// Advance ticke work
}

//
//	UART interrupt handler
//
__declspec(interrupt) void SciHandler(void)
{
UINT8 st, ch;		// Work for status & data
	st = 	MCF_UART0_UISR;	// Get UART interrupt status register
	if(st & MCF_UART_UISR_FFULL_RXRDY)	// Receive data?
	{		// Yes, Serial data has arrived
		ch = MCF_UART0_URB;	// Get arrived data from data buffer
		if(RxRingOut != ((RxRingIn + 1) & RX_MASK))	// Ring buffer has space to put?
		{
			RxBuf[RxRingIn++] =	ch;	// Store arrived data to ring buffer
			if(((RxRingIn - RxRingOut) & RX_MASK) > RX_MASK * 3 / 4)	// Half of ring buffer occupied?
				MCF_UART0_UOP0 = 1;	// Set RTS off for flow control(disable flow)
		}
		return;	// Don't go to transmit handling(another intterrupt happen when needed)
	}
	if(st & MCF_UART_UISR_TXRDY)	// Trnsmitter is ready to send?
	{	// Yes,then send one char and stop when all data is sent
		if(TxRingIn == TxRingOut)	// Transmit ring buffer is empty?
		{	// Yes,no more data to transmit then disable Tx interrupt
			MCF_UART0_UIMR = MCF_UART_UIMR_FFULL_RXRDY;// Clear Tx interrupt bit
			return;	// No more data to send then stop Tx intterrupt and return
		}
		MCF_UART0_UTB = TxBuf[TxRingOut++];	// Send one char from transmit ring buffer
	}
}

//
//	Sense data of UART
//
UINT8 SenseByte(void)
{
	return (UINT8)(RxRingIn != RxRingOut);	// Check ring buffer has data
}

//
//	Get one char from ring buffer
//
UINT8 GetByte(void)
{
UINT8 getc;	// char save work
	while(!SenseByte())SystemSleep();	// Wait until UART data has come
  getc = RxBuf[RxRingOut++];	// Get one char from ring buffer
	if(((RxRingIn - RxRingOut) & RX_MASK) < RX_SIZE / 4)// Data amout is less than 1/4?
		MCF_UART0_UOP1 = 1;		// set RTS on for flow control (flow enable)
	return getc;	// Return with UART input data
}

//
//	Put one char to send queue
//
void PutByte(UINT8 ch)
{
	if(TxRingOut != ((TxRingIn + 1) & TX_MASK))	// Ring buffer has space to put?
	{
		TxBuf[TxRingIn++] = ch;	// Put data to transmit ring buffer
	}
}

//
//	Init timer & UART
//
void InitDevice(UINT8 baud)
{
//	Setup interval timer
	*(UINT32*)0x2000014c = (UINT32)TimerHandler;	// Set timer handler address(interrupt 19)
	MCF_INTC0_ICR19 = MCF_INTC_ICR_IL(5) | MCF_INTC_ICR_IP(5);// Set interrupt priority (5)
	MCF_INTC0_IMRL &= ~MCF_INTC_IMRL_INT_MASK19;	// Clear interrupt 19 mask bit(allow int)
	MCF_DTIM0_DTMR = 0x1a;		// Use internal clock & repeat & enable interrupt
	MCF_DTIM0_DTRR = 60000 - 1;	// Set timer interval to 1ms
	MCF_DTIM0_DTMR |= 1;			// Enable timer module
//	Allocate UART ring buffer for Tx & Rx
	RxBuf = MemoryAlloc(RX_SIZE);	// Allocate Rx ring buffer
	TxBuf = MemoryAlloc(TX_SIZE);	// Allocate Tx ring buffer
	RxRingIn = 0;	// Clear ring in pointer for receive
	RxRingOut = 0;// Clear ring out pointer for receive
	TxRingIn = 0;	// Clear ring in pointer for transmit
	TxRingOut = 0;// Clear ring out pointer for transmit
//	Setup UART registers
	MCF_GPIO_PUAPAR = 0		// Port purpose setting
		| MCF_GPIO_PUAPAR_URXD0_URXD0	// Use Rx pin for UART (not for GPIO)
		| MCF_GPIO_PUAPAR_UTXD0_UTXD0	// Use Tx pin for UART
		| MCF_GPIO_PUAPAR_URTS0_URTS0	// Use RTS pin for UART
		| MCF_GPIO_PUAPAR_UCTS0_UCTS0;// Use CTS pin for UART
	MCF_UART0_UCR = MCF_UART_UCR_RESET_TX;// Reset Transmitter
	MCF_UART0_UCR = MCF_UART_UCR_RESET_RX;// Reset Receiver
	MCF_UART0_UCR = MCF_UART_UCR_RESET_MR;// Reset Mode register write sequence
	MCF_UART0_UMR1 = 0				//	Set serial data format
		| MCF_UART_UMR_PM_NONE	// No parity
		| MCF_UART_UMR_BC_8;		// 8 bit data
	MCF_UART0_UMR2 = 0				//	Access UMR2 after UMR1 write(sequencial)
		| MCF_UART_UMR_SB_STOP_BITS_1;	// 1 stop bit
	MCF_UART0_UCSR = 0				// Set Rx and Tx uses SYSTEM CLOCK
		| MCF_UART_UCSR_RCS_SYS_CLK		// Rx uses system clock
		| MCF_UART_UCSR_TCS_SYS_CLK;	// Tx uses system clock
	MCF_UART0_UIMR = 0				// Allow Rx arrive interrupt
		| MCF_UART_UIMR_FFULL_RXRDY;
	MCF_UART0_UOP1 = 1;				// Set RTS ON to enable flow control
	MCF_UART0_UBG1 = 0;				// Set baud rate register 1 (high byte)
	MCF_UART0_UBG2 = baud;			// Set baud rate register 2 (low byte)
	*(UINT32*)0x20000134 = (UINT32)SciHandler;// Set UART handler address(interrupt 17)
	MCF_INTC0_ICR13 = MCF_INTC_ICR_IL(2);	// Set interrupt priority (2)
	MCF_INTC0_IMRL &= ~MCF_INTC_IMRL_INT_MASK13;	// Clear interrupt 17 mask bit (enable)
	MCF_UART0_UCR = 0					// Finally enable Rx and Tx
		| MCF_UART_UCR_TX_ENABLED		// Set Tx enable bit
		| MCF_UART_UCR_RX_ENABLED;	// Set Rx enable bit
}

//
//	Setup UDP packet
//
void SetupData(void)
{
UINT16	index;
	TxPacketBuffer->AckSeq = RxSeqNum;	// Update received seq num for acknowledge
	if(TxPacketBuffer->Size)		// Last UDP data confirmed?
		return;					// No then re-send same UDP packet (error retry)
	for(index = 0;;)	// Data setup loop
	{
		if(SenseByte() == 0)	// Check UART data has come?
			break;				// No more data then exit setup loop
		TxPacketBuffer->data[index]	= GetByte();	// Store char to packet buffer
		TxPacketBuffer->Size = ++index;	// Advance index and payload size
		if(index == ETHERBUF)	// Pay load size limit?
			break;	// Exit data setup loop when payload is full
	}
	TxPacketBuffer->Size = index;	// Set payload length
	if(++TxSeqNum == 0)	// Increment Packet seq number for new UDP packet
		TxSeqNum++;		// If it's 0 then force to 1
	TxPacketBuffer->SeqNum = TxSeqNum;	// Set Send Seq num
}

//
//	Handle incoming UDP packet
//
void DataIndication(void)
{
UINT16	i;	// Loop counter work
	if(RxPacketBuffer->AckSeq == TxSeqNum)	// Ack seq num matches sent seq num?
		TxPacketBuffer->Size = 0;	// Clear size for next packet send(acknowledged)
	if((MCF_UART0_UIP & 1)	// Check CTS off for flow control
	|| ((TxRingIn - TxRingOut) & TX_MASK) + RxPacketBuffer->Size >= TX_SIZE)
// Check recieved payload will fit Tx ring buffer?(too big payload?)
	{	// Noway to put received data to UART transmit ring buffer
		RxSeqNum = 0; // Force request resend UDP buffer(CTS off or Tx ring is full)
		return;		// Return (not set data to Tx ring buffer)
	}
	RxSeqNum = RxPacketBuffer->SeqNum; // Save recieved packet seq number
	if(RxSeqNum == RxLastNum)	// Already processed last time?
		return;			// Return if it was duplicate packet (dump packet)
	RxLastNum = RxSeqNum;	// Save processed packet seq num
	if(RxPacketBuffer->Size)	// Received packet has payload?
	{	// Yes then put UDP payload to Tx ring buffer for UART transmittion
		for(i = 0; i < RxPacketBuffer->Size; i++)	// Send payload data to Tx ring
			PutByte(RxPacketBuffer->data[i]);	// Put received char to Tx ring buffer
		MCF_UART0_UIMR = MCF_UART_UIMR_FFULL_RXRDY | MCF_UART_UIMR_TXRDY;
// Enable UART tx interrupt and Tx ready interrupt will happen		
	}
}

//
//	Setup parameter
//	IP(4byte),SrcPort(2byte),DstPort(2byte),Baud(1byte),
//	Master or slave(1byte),Interval(1byte)
//
UINT32 Setting(UINT8 *par)
{
	ip = *(UINT32*)par;	// Set ip adrs
	SrcPort = *(UINT16*)(par + 4);	// Get SrcPort
	if(SrcPort == 0)		// Has value?
		SrcPort = 9999;		// Set default port number
	DstPort = *(UINT16*)(par + 6);	// Get DstPort
	if(DstPort == 0)		// Has value?
		DstPort = 9999;		// Set default port number
	Baud = *(par + 8);	// Get baud rate
	if(Baud > 6)				// Baud rate range check
		Baud = 1;					// Set 57600 bps default
	Master = *(par + 9);		// Get Master or Slave mode
	LoopTime = *(par + 10);	// Set main loop time
	return ip;
}

//
//	Network serial cable main
//
void SerialMain(void)
{
//	Setup parameters
	InitDevice(BaudRate[Baud]);	// Init timer & UART and set baud rate
	TxPacketBuffer = MemoryAlloc(sizeof(PacketBuffer_t));	// Alloc UDP send buffer
	MemClear(TxPacketBuffer, sizeof(PacketBuffer_t));
	Soc = CreateSocket(0);	// Create UDP socket
	Bind(Soc, SrcPort, 0);	// Set source port number
// Start main loop
	for(;;)	// Repeat forever
	{	// Transmit UDP and wait UDP reply 
		TickCount = 0;	// Reset time
		SetupData();		// Setup UDP packet
		SendTo(Soc, ip, DstPort,	// Send UDP packet 
			(UINT8*)TxPacketBuffer, (UINT16)(TxPacketBuffer->Size + 5));// Pointer & length
		Adjust = 0;			// Set Default value
		TxPacketBuffer->Time = LoopTime;	// Preset Timeout arrival time 
		while(TickCount < LoopTime - 1)	// Wait reply packet until 1ms before main loop time
		{
			if(RecvFrom(Soc, 0) >= 0)	// Check UDP packet arrival
			{	// Yes,UDP data has arrived
				RxPacketBuffer = (PacketBuffer_t*)GetReceiveBuffer(Soc, 1);// Get UDP data buffer
				if(GetSenderIP(Soc) == ip	// Check IP adrs
					&& GetSenderPort(Soc) == DstPort)	 // Check port number
				{
					TxPacketBuffer->Time = TickCount;	// Reply received time for SYNC
					Adjust = RxPacketBuffer->Time;	// Save time adjust value
					DataIndication();	// Handle incoming UDP packet
					MemoryFree(RxPacketBuffer);	// Release recieved UDP data buffer
				}
				else
					MemoryFree(RxPacketBuffer);	// Dump recieved UDP data buffer
			}
			SystemSleep();	// Allow another thread to execute their task
		}
		if(Master)
		{
			while(TickCount < LoopTime + 1 - Adjust)	// Adjust wait time (only Master)
				SystemSleep();	// Allow another thread to execute their task
		}
		else
		{
			while(TickCount < LoopTime)	// Wait until main loop time
				SystemSleep();	// Allow another thread to execute their task
		}
	}
}
